#!/usr/bin/env python
"""
Set up tagging server on port 2020 and use it to tag a directory of
data.  The idea is to provide an interface to the following calls to
the Staford POS tagger.

SET UP A SERVER

% java -mx300m -cp stanford-postagger-2010-05-26.jar edu.stanford.nlp.tagger.maxent.MaxentTaggerServer -model /Users/jeanmarkgawron/Desktop/SWIDSAS/software/stanford-postagger-full-2010-05-26/models/bidirectional-distsim-wsj-0-18.tagger -port 2020 & 
[1] 405
Loading default properties from trained tagger /Users/jeanmarkgawron/Desktop/SWIDSAS/software/stanford-postagger-full-2010-05-26/models/bidirectional-distsim-wsj-0-18.tagger
Reading POS tagger model from /Users/jeanmarkgawron/Desktop/SWIDSAS/software/stanford-postagger-full-2010-05-26/models/bidirectional-distsim-wsj-0-18.tagger ... done [7.6 sec].

INPUT A FILE TO THE SERVER WAITING ON PORT 2020:

% cat sample-input.txt | java -cp stanford-postagger-2010-05-26.jar edu.stanford.nlp.tagger.maxent.MaxentTaggerServer -client -port 2020

Issues::

  --  Server sometimes hangs or runs out of memory and must be killed
  and restarted on the NEXT File.  Keep a list of the BAD files
  but for now just ignore them and go on.

  --  Also keep a list of files on which the server aborted
  or stopped tagging early.  For example, a stray skipped line
  causes the tagger to conclude the file has ended.  At the moment
  this is tested only by checking whether the output file has
  grown shorter after tagging.

  --  Little client processes may have to be killed too, or memory will
  fill with lots of nasty little java processes. Normally these terminate
  when their requests are completed, but in case of trouble they may linger
  about.

  --  Synchronicity issues. Bandaids in place.  Starting up a
  server takes time and the server does not react well to requests
  piling up while it is still loading its model.  Sleep commands
  inserted for this.  Time for start up currently hard-coded.  No
  known way to know when the server is READY for requests.  
"""


##################################################################
##################################################################
###########   Function   DEFS
##################################################################
##################################################################


import os.path, subprocess, time, sys
## Modules written for this package
import clean_up_text_file, utils, crawling_utils

def make_new_text_file_from_html_file(test_f0):
    (base, ext) = os.path.splitext(test_f0)
    try:
        return crawling_utils.convert_html_file(base+'.html',text_file_extension=ext)
    except IOError:
        (data_dir,filename) = os.path.split(test_f0)
        if filename == '.txt':
            try:
                return crawling_utils.convert_html_file(test_f0 +'.html',text_file_extension=ext,new_base='txt')
            except IOError:
                print 'No html file found (.txt case): %s' % test_f0
        else:
            print 'No html file found: %s' % test_f0

def valid_output_file (in_stat, outfile):
    out_stat = os.stat(outfile)
    if out_stat.st_size == 0:
        return False
    if out_stat.st_size <= in_stat.st_size:
        return 'Redo'
    else:
        return True
    
def update_file_list(file_list_fh, test_f0):
    print >> file_list_fh, test_f0
    file_list_fh.flush()  # stay uptodate in case of crash
   
def restart_server(pid):
    print >> sys.stderr, '******   Restarting Server! *******'
    kill_server(pid)
    return start_server()

def start_server ():
    P = subprocess.Popen(os.path.join(stanford_app_dir,\
                                      'start_stanford_pos_server'), \
                         shell=True)
    print >> sys.stderr, 'Tagging Server started', P.pid + 1
    ## It takes this puppy some time toload its model.
    time.sleep(7)
    # Why 1 must be added I'll never know.  But the OS seems to think so.
    return P.pid + 1

def kill_server (pid):
    os.system('kill -9 %s' % pid)


##################################################################
##################################################################
###########   Global VARS
##################################################################
##################################################################    

# tag_server_command_str = 'java -mx300m -cp stanford-postagger-2010-05-26.jar edu.stanford.nlp.tagger.maxent.MaxentTaggerServer -model /Users/jeanmarkgawron/Desktop/SWIDSAS/software/stanford-postagger-full-2010-05-26/models/bidirectional-distsim-wsj-0-18.tagger -port 2020 &'

# tag_command_str0 = 'java -mx750m -classpath stanford-postagger.jar edu.stanford.nlp.tagger.maxent.MaxentTagger -model models/bidirectional-distsim-wsj-0-18.tagger -textFile %s > %s'

tag_command_str0 = "java -cp '%s/stanford-postagger-2010-05-26.jar:' edu.stanford.nlp.tagger.maxent.MaxentTaggerServer -client -port 2020"


##################################################################
##################################################################
###########   Script body
##################################################################
##################################################################

try:

    #####  DATA  ###################################
    ### Define VAR C{rel_data_dir} where listing of
    ### files to be tagged can be found.  This listing
    ### has been usually created wih a Unix C{find} command
    ### and paths are all relative to the C{rel_ata_dir}
    
    home = os.environ['HOME']
    rel_data_dir = os.path.join(home, \
                                'Desktop/SWIDSAS/data')
    index_file = os.path.join(rel_data_dir, 'website_files.listing')
    print 'Using index file %s' % index_file

    untagged_ext = '.txt'
    tagged_ext = '.tagged'

    
    #####   STANFORD TAGGER DEFS  ##########
    ### Define VAR stanford_app_dir #####################
    app_root_dir = os.path.join(home,'Desktop','SWIDSAS','software')
    # print 'Root: ', app_root_dir
    rel_stanford_app_dir = 'stanford-postagger-full-2010-05-26'
    stanford_app_dir = os.path.join(app_root_dir, rel_stanford_app_dir)
    # print 'Stanford app dir: ', stanford_app_dir
    # os.chdir(stanford_app_dir)

    ###  OUTPUT DIR #######################
    ### Define VAR out_dir (for tagged files) #####################
    # cur_dir = os.getcwd()
    # out_dir = cur_dir
    out_dir = os.path.join(rel_data_dir,'clean_data')
    # print 'Output dir: ', out_dir
    
    ###############################################################
    ###########  Tagging  #########################################
    ###############################################################

    ## Info for resuming after interruption
    (donefiles, done_files_fh) = utils.get_filestatus_info('done', rel_data_dir)
    (badfiles, bad_files_fh) = utils.get_filestatus_info('bad', rel_data_dir)
    (redofiles, redo_files_fh) = utils.get_filestatus_info('redo', rel_data_dir)
    ## This one just gets read.  No updating.
    index_fh = open(index_file,'r')
    ## start up Tagging server
    pid0 = start_server()
    
    # print 'starting server: ', tag_server_command_str
    #(p_stdout, p_stdin) = os.popen2(tag_server_command_str)
    #print p_stdout

    ##  
    ctr = 0
    ## Change to location all the filepaths in C{index_file} are relative to.
    os.chdir(rel_data_dir)
    ## Accumulator vars for child client tagging processes
    ## These could be probably be kill upon completion. But
    ##  we are currently buffering and killing a batch of them
    ## of them at a time.
    (old_pid_list, pid_list) = ([ ], [ ])
    for test_f0 in index_fh:
        test_f0 = test_f0.strip()
        if not os.path.isfile(test_f0):
            continue
        if test_f0 in donefiles:
            continue
        if test_f0 in badfiles:
            continue
        if test_f0 in redofiles:
            continue
        (base, ext) = os.path.splitext(test_f0)
        if not base or base == '.':
            continue
        if ext <> untagged_ext:
            continue
        #  To repair some bad decisions in file cleanup
        #  We are modifying the file test_f0 before going on.
        new_test_f0_base = make_new_text_file_from_html_file(test_f0)
        if new_test_f0_base:
            new_test_f0 = new_test_f0_base + untagged_ext
        else:
            ## werent able to redo it. Try tagging the original version.
            new_test_f0 = test_f0
        chunk_list = clean_up_text_file.clean_and_chunk(new_test_f0)
        for test_f in chunk_list: ## Allowing chunk_list to be a list to break up long files.
            (base, ext) = os.path.splitext(test_f)
            if base:
                ###  Construct name for outfile,
                ###  using components of path to infile
                ###  Skip uniformative root_dir ("website_files")
                tagged_out_file_name = '__'.join(base.split(os.path.sep)[1:])
                ## Add extensions
                tagged_out_file_name = tagged_out_file_name+tagged_ext+ext
                file_to_tag = os.path.join(rel_data_dir,\
                                           test_f)
                in_stat = os.stat(file_to_tag)
                tagged_file = os.path.join(out_dir, \
                                           tagged_out_file_name)
                if os.path.isfile(file_to_tag):
                    print
                    print >> sys.stderr, 'File to tag: %s' % test_f
                    print >> sys.stderr, 'Tagged out: %s' % tagged_file
                    print >> sys.stderr
                    fh = open(file_to_tag,'r')
                    fh_tagged = open(tagged_file,'w')
                    P1 = subprocess.Popen(tag_command_str0 % stanford_app_dir,stdin=fh,stdout=fh_tagged,shell=True,universal_newlines=True)
                    # print tag_command_str0 % stanford_app_dir
                    pid = P1.pid
                    # time.sleep(1)
                    os.waitpid(pid,0)
                    fh.close()
                    fh_tagged.close()
                    tagging_succeeded = valid_output_file(in_stat,tagged_file)
                    if tagging_succeeded:
                        if tagging_succeeded == 'Redo':
                            update_file_list(redo_files_fh, test_f0)
                        else:
                            update_file_list(done_files_fh, test_f0)
                    else:
                        os.remove(tagged_file)  # Dont clutter data with empty files.
                        update_file_list(bad_files_fh, test_f0)
                        pid0 = restart_server(pid0)
                    ctr += 1
                    pid_list.append(pid)
                    if ctr%50 == 0:
                        for p in old_pid_list:
                            os.system('kill -9 %s' % p)
                        old_pid_list = pid_list
                        pid_list = []
                        print >> sys.stderr, '******* %d files tagged *******' % (ctr,)
    done_files_fh.close()
    bad_files_fh.close()
    redo_files_fh.close()
    index_fh.close()
    kill_server(pid0)

except:

    raise

finally:

    done_files_fh.close()
    bad_files_fh.close()
    redo_files_fh.close()
    index_fh.close()
    kill_server(pid0)
    
